/**CCSA.cpp***********************************************************************************************
 *          Copyright 2004, V. Makinen                                                                   *
 *                                                                                                       *
 * This file contains implementation of the compressed compact suffix array (CCSA).                      *
 * For more information, see [MN04].                                                                     *                     
 *                                                                                                       *
 *                                                                                                       *   
 * This software may be used freely for any non-commercial purpose. However,                             *
 * when distributed, the original source must be clearly stated, and, when the                           *
 * source code is distributed, the copyright notice must be retained and any                             *
 * alterations in the code must be clearly marked. No warranty is given regarding                        *
 * the quality of this software.                                                                         *
 *                                                                                                       *
 * [MN04] Veli Mkinen and Gonzalo Navarro. Compressed Compact Suffix Arrays.                            *
 *        To appear in CPM 2004                                                                          *
 *                                                                                                       *
 *********************************************************************************************************/


//---------------------------------------------------------------------------
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/times.h>
#include <assert.h>
#include <string.h>
#include "qsufsort.c"
//---------------------------------------------------------------------------

#define W 32 // computer word size
#define Wminusone 31  // W-1
#define WW 64  // 2*W

unsigned n,m; // text length, pattern length
bool print=true; 
int report=1;  // 0=existence, 1=report, 2=count, X>2 = print X characters after each pattern occurrence
unsigned bitsn=0; // bits needed to represent n

char* text; 

class TSA;
class TCCSA;
TSA *CreateSuffixArray(char *text, unsigned n);
TCCSA *MakeSuffixArrayAndCompressIt(char *text, unsigned n);


/**********************************************************************************************************
 * Some preliminary stuff                                                                                 *
 **********************************************************************************************************/

#define mask31 0x0000001F

#define max(x,y) ((x)>(y)?(x):(y))
#define min(x,y) ((x)<(y)?(x):(y))


#define W 32 //numero de bits del entero de la maquina
#define Wminusone 31  // W-1
#define WW 64 //numero de bits del entero de la maquina
#define bitsM 8 //bits para hacer la mascara para contar mas rapido
#define BW 4 //bytes que hacen una palabra

const unsigned char __popcount_tab[] =
{
0,1,1,2,1,2,2,3,1,2,2,3,2,3,3,4,1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,4,5,5,6,5,6,6,7,5,6,6,7,6,7,7,8,
};

const unsigned char select_tab[] =
{
0,1,2,1,3,1,2,1,4,1,2,1,3,1,2,1,5,1,2,1,3,1,2,1,4,1,2,1,3,1,2,1,

6,1,2,1,3,1,2,1,4,1,2,1,3,1,2,1,5,1,2,1,3,1,2,1,4,1,2,1,3,1,2,1,

7,1,2,1,3,1,2,1,4,1,2,1,3,1,2,1,5,1,2,1,3,1,2,1,4,1,2,1,3,1,2,1,

6,1,2,1,3,1,2,1,4,1,2,1,3,1,2,1,5,1,2,1,3,1,2,1,4,1,2,1,3,1,2,1,

8,1,2,1,3,1,2,1,4,1,2,1,3,1,2,1,5,1,2,1,3,1,2,1,4,1,2,1,3,1,2,1,

6,1,2,1,3,1,2,1,4,1,2,1,3,1,2,1,5,1,2,1,3,1,2,1,4,1,2,1,3,1,2,1,

7,1,2,1,3,1,2,1,4,1,2,1,3,1,2,1,5,1,2,1,3,1,2,1,4,1,2,1,3,1,2,1,

6,1,2,1,3,1,2,1,4,1,2,1,3,1,2,1,5,1,2,1,3,1,2,1,4,1,2,1,3,1,2,1,
};


// bits needed to represent a number between 0 and n
inline uint bits (uint n){
	uint b = 0;
	while (n) { b++; n >>= 1; }
	return b;
}

// reads bit p from e
#define bitget(e,p) ((((e)[(p)/W] >> ((p)%W))) & 1)
// sets bit p in e
#define bitset(e,p) ((e)[(p)/W] |= (1<<((p)%W)))
// cleans bit p in e
#define bitclean(e,p) ((e)[(p)/W] &= ~(1<<((p)%W)))

// numero de enteros necesarios para representar e elementos de largo n
#define enteros(e,n) ((e)*(n))/W+(((e)*(n))%W > 0)

inline unsigned GetField(unsigned *A, register  unsigned len, register unsigned index) {
register unsigned i=index*len/W, j=index*len-W*i, result;
if (j+len <= W)
result = (A[i] << (W-j-len)) >> (W-len);
else {
result = A[i] >> j;
result = result | (A[i+1] << (WW-j-len)) >> (W-len);
}
return result;
};

inline unsigned GetFieldW32(unsigned *A,register unsigned index) {
return A[index];
};

inline unsigned GetFieldW16(unsigned *A,register unsigned index) {
register unsigned i=index/2, j=(index&1)<<4, result;
result = (A[i] << (16-j)) >> (16);
return result;
};

inline unsigned GetFieldW4(unsigned *A,register unsigned index) {
register unsigned i=index/8, j=(index&0x7)<<2;
//register unsigned i=index/8, j=index*4-32*i;
return (A[i] << (28-j)) >> (28);
};

inline void SetField(unsigned *A,register unsigned len, register unsigned index,register  unsigned x) {
   unsigned i=index*len/W, 
            j=index*len-i*W;
   unsigned mask = (j+len < W ? ~0u << j+len : 0) 
                   | (W-j < W ? ~0u >> (W-j) : 0);
   A[i] = (A[i] & mask) | x << j;
   if (j+len>W) {
      mask = ((~0u) << (len+j-W));  
      A[i+1] = (A[i+1] & mask)| x >> (W-j);
   } 
};

inline void SetField32(unsigned *A, register unsigned index,register unsigned x) {
   A[index]=x;
}; 


inline uint popcount (register int x){
  return __popcount_tab[(x >>  0) & 0xff]  + __popcount_tab[(x >>  8) & 0xff]  + __popcount_tab[(x >> 16) & 0xff] + __popcount_tab[(x >> 24) & 0xff];
};

inline uint popcount16 (register int x){
  return __popcount_tab[x & 0xff]  + __popcount_tab[(x >>  8) & 0xff];
};

inline uint popcount8 (register int x){
  return __popcount_tab[x & 0xff];
};

/**********************************************************************************************************
 * Suffix array in about n log n bits                                                                     *
 **********************************************************************************************************/

class TSA {
public:
    unsigned posbits;
    unsigned *pos;
    unsigned *B;    
    unsigned n;
public:
    void CreatePosArray(unsigned bits, unsigned n) 
      {this->n = n; posbits = bits; pos = new unsigned[n*posbits/W+2];
       CreateBooleanArray(n);};
    void CreateBooleanArray(unsigned n)
      {B = new unsigned[n/W+2];};   
    void DeletePosArray() {delete pos;};
    void DeleteBooleanArray() {delete B;};        
    void SetPos(unsigned index, unsigned x)
      {SetField(pos,posbits,index,x);};
    void SetB(unsigned index, unsigned x)
      {SetField(B,1,index,x);};
    unsigned GetPos(unsigned index)
      {return GetField(pos,posbits,index);};        
    bool GetB(unsigned index)
      {return GetField(B,1,index);};        
    ~TSA() {DeletePosArray(); DeleteBooleanArray();};
};

TSA *CreateSuffixArray(char *text, unsigned n)
{
   TSA *SA;
   int *x, *p;
   p=(int*)malloc((n+1)*sizeof *p);
   x=(int*)malloc((n+1)*sizeof *x);
   unsigned i;
   for (i=0; i<n; i++)
      x[i]=(int)text[i]+128;
   x[n]=0;
   suffixsort(x,p,n,256,0);
   free(x);
   unsigned k=0;
   while ((n<<k)>>(W-1) == 0) k++;
   bitsn = W-k;
   SA = new TSA();
   SA->CreatePosArray(bitsn,n);
   for (i=0; i<n; i++) {
      SA->SetPos(i,p[i+1]);
      SA->SetB(i,0);
   }
   free(p);
   return SA;
}

/**********************************************************************************************************
 * New RANK and SELECT implementations from Rodrigo Gonzales, Univ. Chile                                 *
 * I modified the code a bit adding some more useful functions                                            * 
 **********************************************************************************************************/

class BitRankF {
private:
	uint *data; //here is the bit-array
	bool owner;
	uint n,integers;
	uint factor,b,s; 
	uint *Rs; //superblock array
	unsigned char *Rb; //block array

	uint BuildRankSub(uint ini,uint fin); //internal use of BuildRank
	int BuildRank(); //crea indice para rank
public:
	BitRankF(uint *bitarray, uint n, bool owner);
	~BitRankF(); //destructor	 
	bool IsBitSet(uint i);
	uint rank(uint i); //Rank from 0 to n-1
	uint prev(uint start); // gives the largest index i<=start such that IsBitSet(i)=true 
        uint select(uint x); // gives the position of the x:th 1.
        uint SpaceRequirementInBits();	
};

/////////////
//Rank(B,i)// 
/////////////
//This Class use a superblock size of 256 bits
//and a block size of 32 bits also
//we use a 32 bit to represent Rs and 
//we use a 8 bit to represent Rb

BitRankF::BitRankF(uint *bitarray, uint n, bool owner) {
	data=bitarray;
	this->owner = owner;
	this->n=n;
	b=32; // b is a word
	factor=8; // 8 word in 256 bits
	s=b*factor;
	uint aux=(n+1)%W;
	if (aux != 0)
		integers = (n+1)/W+1;
	else 
		integers = (n+1)/W;
	BuildRank();
}

BitRankF::~BitRankF() {
	delete [] Rs;
	delete [] Rb;
	if (owner) delete data;
}

//Build the rank (blocks and superblocks)
int BitRankF::BuildRank(){
	uint num_sblock = n/s;
	uint num_block = n/b;
	Rs = new uint[num_sblock+1];//+1 we add the 0 pos
	Rb = new unsigned char[num_block+1];//+1 we add the 0 pos
	int j;
	Rs[0]=0;
	for (j=1;j<=num_sblock;j++) 
		Rs[j]=BuildRankSub((j-1)*factor,factor)+Rs[j-1];
	
	Rb[0]=0;
	for (int k=1;k<=num_block;k++) {
		j = k / factor;
		Rb[k]=BuildRankSub(j*factor,k%factor);
	}
}

uint BitRankF::BuildRankSub(uint ini,uint bloques){
	uint rank=0,aux;
	for(uint i=ini;i<ini+bloques;i++) {
		if (i < integers) {
			aux=data[i];
			rank+=popcount(aux);
		}
	}
	return rank; //return the numbers of 1's in the interval
}


//this rank ask from 0 to n-1
uint BitRankF::rank(uint i) {
	++i; // the following gives sum of 1s before i 
	return Rs[i>>8]+Rb[i>>5]
		+popcount(data[i>>5] & ((1<<(i & mask31))-1));
}

uint BitRankF::prev(uint start) {
      // returns the position of the previous 1 bit before and including start.
      // tuned to 32 bit machine

      unsigned i = start >> 5;
      int offset = (start % W);
      unsigned answer = start;
      unsigned int val = data[i] << (Wminusone-offset);

      if (!val) { val = data[--i]; answer -= 1+offset; }

      while (!val) { val = data[--i]; answer -= W; }

      if (!(val & 0xFFFF0000)) { val <<= 16; answer -= 16; }
      if (!(val & 0xFF000000)) { val <<= 8; answer -= 8; }

      while (!(val & 0x80000000)) { val <<= 1; answer--; }
      return answer;
}

uint BitRankF::select(uint x) {
	// returns i such that x=rank(i) && rank(i-1)<x or n if that i not exist
	// first binary search over first level rank structure
	// then sequential search using popcount over a int
	// then sequential search using popcount over a char
	// then sequential search bit a bit
	
	//binary search over first level rank structure
	unsigned l=0, r=n/s;
	unsigned mid=(l+r)/2;      
	unsigned rankmid = Rs[mid];
	while (l<=r) {
		if (rankmid<x)
			l = mid+1;
		else
			r = mid-1;
		mid = (l+r)/2;              
		rankmid = Rs[mid];
	}    
	//sequential search using popcount over a int
	unsigned left;
	left=mid*factor;
	x-=rankmid;
        uint j=data[left];
        uint ones = popcount(j);
        while (ones < x) {
	  x-=ones;left++;
	  if (left > integers) return n;
          j = data[left];
  	  ones = popcount(j);
        }
	//sequential search using popcount over a char
	left=left*b; 
	rankmid = popcount8(j);
	if (rankmid < x) {
		j=j>>8;
		x-=rankmid;
		left+=8;
		rankmid = popcount8(j);
		if (rankmid < x) {
			j=j>>8;
			x-=rankmid;
			left+=8;
			rankmid = popcount8(j);
			if (rankmid < x) {
				j=j>>8;
				x-=rankmid;
				left+=8;
			}
		}
	}

	// then sequential search bit a bit
        while (x>0) {
		if  (j&1) x--;
		j=j>>1;
		left++;
	}
	return left-1;
}

bool BitRankF::IsBitSet(uint i) {
	return (1u << (i % W)) & data[i/W];
}

uint BitRankF::SpaceRequirementInBits() {
	return (owner?n:0)+W*n/s+bitsM*n/b;
}

/**********************************************************************************************************
 * Compressed Compact Suffix Array implementation                                                         *
 **********************************************************************************************************/
 
class TCCSA {
public:
    unsigned posbits;
    unsigned *pos;
    unsigned length; 
    int *chars;
    BitRankF *charrank;
    BitRankF *posrank;
    BitRankF *samplerank;
    unsigned *sample;    
    unsigned samplebits;
    unsigned samplerate; // suffixes 0,samplerate,2*samplerate,...., are sampled  
    unsigned samplen;
public:
    void CreatePosArray(unsigned bits, unsigned n) 
      {posbits = bits; pos = new unsigned[(n*bits)/W+1];};
    void DeletePosArray() {delete pos;};
    void SetPos(unsigned index, unsigned x)
      {SetField(pos,posbits,index,x);};
    unsigned GetPos(unsigned index)
      {return GetField(pos,posbits,index);};        
    inline unsigned GetLen(unsigned index)
      {return posrank->select(index+2)-posrank->select(index+1);}; 
    void SetSample(bool *B, unsigned n, unsigned *sample, unsigned samplebits, unsigned samplen, unsigned samplerate) {
       this->sample = sample; this->samplebits = samplebits; this->samplen = samplen;
       this->samplerate = samplerate;
       unsigned i; 
       unsigned *Binbits = new unsigned[n/W+1];
       for (i=0;i<n;i++)
          SetField(Binbits,1,i,B[i]);       
       samplerank = new BitRankF(Binbits,n,true);
       //samplerank->Test();
       // the caller can delete B
    };    
    void CreateRankForChars(char *text, unsigned int n) { 
       int i,j=0,sumlen=0,sumchars=0;
       chars = new int[256]; // chars[c+128] tells how many times character c repeats
       char *charatindex = new char[length]; // charatindex[i] tells which char is at entry i
       for (i=0;i<256; i++) chars[i]=0;
       for (i=0;i<n;i++)
          chars[text[i]+128]++; 
       for (i=0;i<length;i++) {
          sumlen += GetLen(i);    
          while (sumlen>sumchars+chars[j])
             sumchars += chars[j++];       
          charatindex[i]=j-128;
       }   
       bool *B=new bool[length];
       j=0; B[0]=1;
       chars[j++]=charatindex[0]; // chars[j] tells which is the jth distinct char
       for (i=1;i<length;i++) {
          if (charatindex[i]!=charatindex[i-1]) {
             B[i] = 1;
             chars[j++] = charatindex[i];              
          } else B[i]=0;       
       }                
       delete charatindex;
       unsigned *Binbits = new unsigned[length/W+1];
       for (i=0;i<length;i++)
          SetField(Binbits,1,i,B[i]);
       charrank = new BitRankF(Binbits,length,true);        
       //charrank->Test();
       delete B;      
    };   
    void CreateRankForPos(bool *B) {
       unsigned i;
       unsigned *Binbits = new unsigned[n/W+1];
       for (i=0;i<n;i++)
          SetField(Binbits,1,i,B[i]);
       posrank = new BitRankF(Binbits,n,true);
       //posrank->Test();
       // caller can delete B
    };
    inline bool IsSampled(unsigned x,unsigned offset) {
       return samplerank->IsBitSet(posrank->select(x+1)+offset);  
    };        
    inline unsigned GetTextPos(unsigned i) {
       unsigned depth=0;
       unsigned x;
       while (!samplerank->IsBitSet(i)) {
          x = posrank->rank(i)-1;         
          i = GetPos(x)+i-posrank->prev(i); //+posrank->select(x+1);                          
          depth++;         
       }
       return GetField(sample,samplebits,samplerank->rank(i)-1)-depth; 
    }; 
    void ShowPath(unsigned i, unsigned len) {
       int x,offset,j; char c;        
       x = posrank->rank(i)-1; // CCSA position
       offset = i-posrank->prev(i); // offset = i-posrank->select(x+1);       
       for (j=0;j<len;j++) {
          c = chars[charrank->rank(x)-1]; //CharAtIndex(x) 
          if (c=='\n') printf("%c",'/');
          else printf("%c",c);
          i = GetPos(x)+offset; // suffix array position
          x = posrank->rank(i)-1; // CCSA position
          offset = i -posrank->prev(i);
       }        
       printf("\n");
    };
    inline char CharAtIndex(unsigned i) { 
       return chars[charrank->rank(i)-1];
    };    
    inline int CompareIndex(char *pattern, unsigned int m, unsigned i) { // i is suffix array position
       // compares pattern[0...m-1] with the suffix
       int x,offset,j; char c;
       x = posrank->rank(i)-1; // CCSA position
       offset = i-posrank->prev(i); // offset = i-posrank->select(x+1);
       for (j=0;j<m;j++) {
          c = chars[charrank->rank(x)-1]; //CharAtIndex(x)
          if (pattern[j]>c) return j+1;
          else if (pattern[j]<c) return -(j+1);
          /* c==pattern[j] */
          i = GetPos(x)+offset; // suffix array position
          x = posrank->rank(i)-1; // CCSA position
          offset = i -posrank->prev(i);
       }
       return 0;  // match
    };

    inline unsigned ReportOccurrenceRegion(unsigned left, unsigned right, unsigned depth) {
       // sa indexes left and right
       unsigned r=right, prevsample=samplerank->prev(right), prevlink=posrank->prev(right);
       unsigned x,y,i;
       while (prevsample >= left || prevlink >= left) {
          if (prevsample < prevlink) {
             // recursively report area prevlink...r
             x = posrank->rank(prevlink)-1;
             i = GetPos(x);
             ReportOccurrenceRegion(i,i+r-prevlink, depth+1);
             r = prevlink -1;
             prevlink = posrank->prev(r);
          }
          else {
             // recursively report area prevsample+1...r
             // report suffix of index prevsample
             if (prevsample<r) {
                i = prevsample+1;
                x = posrank->rank(i)-1;
                i = GetPos(x) + i- prevlink;
                ReportOccurrenceRegion(i,i+r-prevsample-1, depth+1);
             }
             y = GetField(sample,samplebits,samplerank->rank(prevsample)-1)-depth;
             if (print) {
                printf("Match at %d",y);
                fflush(stdout);
                if (report>2) {
                   printf(": ");
                   ShowPath(prevsample,m+report);
                }
                else
                   printf("\n");
             }
             r = prevsample -1;
             prevsample = samplerank->prev(r);
          }
       }
       if (r >= left) {
          // recursively report area left...r
          x = posrank->rank(left)-1;
          i = GetPos(x) + left-posrank->prev(left);
          ReportOccurrenceRegion(i,i+r-left, depth+1);
       }
    };

    inline unsigned ReportOccurrence(unsigned i, unsigned *path, unsigned *pathdepth) {
       unsigned depth=0;
       unsigned x;
       while (*pathdepth > depth && !samplerank->IsBitSet(i) && !posrank->IsBitSet(i))
          i = ++(path[depth++]);
       while (!samplerank->IsBitSet(i)) {
          x = posrank->rank(i)-1;
          i = GetPos(x)+i-posrank->prev(i); //posrank->select(x+1);
          path[depth++]=i;
       }
       *pathdepth = depth;
       return GetField(sample,samplebits,samplerank->rank(i)-1)-depth;
    };

    void ReportOccurrences(unsigned l, unsigned r) {
       unsigned *path = new unsigned[samplerate+1];
       unsigned pathdepth = 0;
       unsigned i, x;
       for (i=l;i<=r;i++) {
          x = ReportOccurrence(i,path,&pathdepth);
          if (print) {
             printf("Match at %d",x);
             fflush(stdout);
             if (report>2) {
                printf(": ");
                ShowPath(i,m+report);
             }
             else
                printf("\n");
          }
       }
       delete path;
     };
     int Search(char *pattern, unsigned int m) {
        int L;
        int R;
        int i;
        int Llimit;
        int Rlimit=n;
        unsigned matches=0;
        // search for left limit
        L = -1;
        R = n;
        i = (L+R)/2;
        while (L<R-1) {
           if (CompareIndex(pattern,m,i) <= 0)
              R = i;
           else
              L = i;
           i = (R+L)/2;
        }
        Llimit=R;
        // search for right limit
        L = Llimit-1;
        R = n;
        i = (L+R)/2;
        if (Llimit < n) {
           while (L < R-1) {
              if (CompareIndex(pattern,m,i) >= 0)
                 L = i;
              else
                 R = i;
              i = (R+L)/2;
           }
           Rlimit=L;
        }
        if (Llimit == n || Llimit > Rlimit) {
            if (print) printf("Pattern (%s) not found\n",pattern);
          return 0; 
        }
        matches=Rlimit-Llimit+1;
        if (!report) {
           if (print && matches) 
              printf("Pattern (%s) found\n",pattern);              
           else if (!matches && print)
              printf("Pattern (%s) not found\n",pattern);
           return matches>0;
        }
        if (report == 2) return matches; // counting
        //ReportOccurrenceRegion(Llimit,Rlimit,0);
        ReportOccurrences(Llimit,Rlimit);
        // above is 25% faster than this:
        /*
        for (L=Llimit;L<=Rlimit;L++) {
           i = GetTextPos(L);
           if (print) {
              printf("Match at %d",i);
              if (report>2) {
                 printf(": ");
                 ShowPath(L,m+report); // report > 2 == how large context is shown
              }
              else
                 printf("\n");
           }
        } */

        return matches;
     };
     void Print(char *text) {
        int i;
        printf("Compressed Compact Suffix Array of text \n %s \n is:\n",text);
        fflush(stdout);
        for (i=0;i<length;i++) {
           printf("%d: (%d,%d) - %c\n",i,GetPos(i),GetLen(i),CharAtIndex(i)); 
           fflush(stdout);
        }
     };     
      
     unsigned SpaceRequirementInBits() {
        return charrank->SpaceRequirementInBits()+ //charrank
                256*W+ // chars
                posrank->SpaceRequirementInBits()+ // posrank
                samplebits*samplen+samplerank->SpaceRequirementInBits()+ // samplerank
                posbits*length; // the array itself
     };
      
     ~TCCSA() {
        delete pos; delete sample; delete chars; 
        delete charrank; delete posrank; delete samplerank;
     };

};

unsigned MarkLinksToSuffixArray(TSA *SArray, TSA *DArray, char *text, unsigned n)
/* This is for making compressed compact suffix arrays */
{
   unsigned newsize=n;
   unsigned i;
   unsigned j;
   unsigned x;
   unsigned index;
   // Make DArray the inverse of SArray
   for (i=0; i<n; i++) {
      DArray->SetPos(SArray->GetPos(i),i);
      DArray->SetB(i,0);
      SArray->SetB(i,0);
   }
   // Swap SArray and DArray
   for (i=0; i<n; i++) {
      x = DArray->GetPos(i);
      DArray->SetPos(i,SArray->GetPos(i));
      SArray->SetPos(i,x);
   }
   i=0;
   /* Mark links */
   while (i<n) {
      x = SArray->GetPos((DArray->GetPos(i)+1) % n);
      DArray->SetB(i,0);
      unsigned textpos=DArray->GetPos(i);
      DArray->SetPos(i,x);
      j = 1;
      while (x+j <n && i+j <n &&                
             SArray->GetPos(((DArray->GetPos(i+j)+1) % n)) == x+j && 
             text[DArray->GetPos(i+j)] == text[textpos]) { // restrict a link to contain only one character            
         DArray->SetB(i+j,1);
         DArray->SetPos(i+j,x+j);
         newsize--;
         j++;
      }
      i = i+j;
   }
   // SArray := inverse of SArray
   for (i=0; i<n; i++) {
      j = SArray->GetPos(i);
      x = i;
      while (SArray->GetB(j)==0) {
         index = j;
         j = SArray->GetPos(index);
         SArray->SetPos(index,x);
         SArray->SetB(index,1);
         x = index;
      }
   } 
   return newsize;
}

void FromMarkedSAToCCSA(TSA *SArray, TCCSA *DArray)
{
   unsigned i=0;
   unsigned j=0;
   bool *B = new bool[n];
   while (i<n) {
      DArray->SetPos(j++,SArray->GetPos(i));
      B[i++]=1;
      while (i<n && SArray->GetB(i))         
        B[i++]=0;    
   }
   B[i]=1;
   DArray->CreateRankForPos(B);
   delete B;
}


TCCSA *MakeCompressedCompactSuffixArray(TSA *SA, char *text, unsigned n)
{
   unsigned newsize=n,i;
   unsigned samplerate=bits(n);
   unsigned samplebits=samplerate;
   unsigned *sample = new unsigned[n/W+2];
   unsigned samplen = n/samplerate+1;
   bool *B = new bool[n];
   unsigned rank=0;
   /* Make sample of suffixes for reporting */
   for (i=0;i<n;i++)
      if (SA->GetPos(i) % samplerate == 0 || SA->GetPos(i) == n-1) {
        SetField(sample,samplebits,rank,SA->GetPos(i));
        B[i]=1;            
        rank++;
      } else B[i]=0;       
   TSA *SATemp = new TSA();
   SATemp->CreatePosArray(bitsn,n);   
   newsize = MarkLinksToSuffixArray(SA, SATemp, text, n); 
   delete SA;
   TCCSA *CCSA = new TCCSA();
   CCSA->length = newsize;
   CCSA->CreatePosArray(bitsn,newsize+1);
   FromMarkedSAToCCSA(SATemp,CCSA);
   delete SATemp;
   CCSA->CreateRankForChars(text,n);
   CCSA->SetSample(B,n,sample,samplebits,samplen,samplerate);
   delete B;
   // not needing text anymore
   //delete text;
   printf("Compressed Compacted Suffix Array is of length %d \n and takes %f B space \n",
          CCSA->length,CCSA->SpaceRequirementInBits()/8.0);
   fflush(stdout);
   return CCSA;
}

TCCSA *MakeSuffixArrayAndCompressIt(char *text, unsigned n)
{
   TSA *SA = CreateSuffixArray(text,n);
   printf("Suffix Array takes %d*%f B = %f B space \n",
          n,(1.0*SA->posbits)/8,
          n*((1.0*SA->posbits)/8));
   printf("Text size is %d B\n",n);
   return MakeCompressedCompactSuffixArray(SA,text,n);    
}

